/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998, 1999, 2000, 2001  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    slouken@devolution.com
*/

/*
    SDL_epocvideo.cpp
    Epoc based SDL video driver implementation

    Epoc version by Hannu Viitala (hannu.j.viitala@mbnet.fi)
*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

extern bool gViewVisible;

extern "C"
{
#include "SDL_error.h"
#include "SDL_timer.h"
#include "SDL_video.h"
#undef NULL
#include "SDL_pixels_c.h"
};

#include "SDL_epocvideo.h"
#include "SDL_epocevents_c.h"

#include <hal.h>
#include <coedef.h>

/* For debugging */
TBool gHeapIsLocked = EFalse;

void RDebug_Print_b(char* error_str, void* param)
    {
    TBuf8<128> error8((TUint8*)error_str);
    TBuf<128> error;
    error.Copy(error8);
    if (param) //!! Do not work if the parameter is really 0!!
        RDebug::Print(error, param);
    else 
        RDebug::Print(error);
    }

extern "C" void RDebug_Print(char* error_str, void* param)
    {
    RDebug_Print_b(error_str, param);
    }

int Debug_AvailMem2()
    {
    //User::CompressAllHeaps();
    TMemoryInfoV1Buf membuf; 
    User::LeaveIfError(UserHal::MemoryInfo(membuf));
    TMemoryInfoV1 minfo = membuf();
	return(minfo.iFreeRamInBytes);
    }

extern "C" int Debug_AvailMem()
    {
    return(Debug_AvailMem2());
    }


extern "C"
{

/* Initialization/Query functions */

static int EPOC_VideoInit( _THIS, SDL_PixelFormat *vformat );

static SDL_Rect **EPOC_ListModes( 
                    _THIS,
                    SDL_PixelFormat *format,
                    Uint32 flags );

static SDL_Surface *EPOC_SetVideoMode(
        _THIS,
        SDL_Surface *current,
        int width,
        int height,
        int bpp,
        Uint32 flags );

static int EPOC_SetColors(
            _THIS,
            int firstcolor,
            int ncolors,
            SDL_Color *colors );

static void EPOC_VideoQuit( _THIS );

/* Hardware surface functions */

static int EPOC_AllocHWSurface( _THIS, SDL_Surface *surface );
static int EPOC_LockHWSurface( _THIS, SDL_Surface *surface );
static int EPOC_FlipHWSurface( _THIS, SDL_Surface *surface );
static void EPOC_UnlockHWSurface( _THIS, SDL_Surface *surface );
static void EPOC_FreeHWSurface( _THIS, SDL_Surface *surface );
static void EPOC_DirectUpdate( _THIS, int numrects, SDL_Rect *rects );

static int EPOC_Available( void );

static SDL_VideoDevice *EPOC_CreateDevice( int devindex );

/* Mouse functions */

static WMcursor *EPOC_CreateWMCursor( 
                    _THIS,
                    Uint8 *data,
                    Uint8 *mask,
                    int w,
                    int h,
                    int hot_x,
                    int hot_y);

static void EPOC_FreeWMCursor( _THIS, WMcursor *cursor );
static int EPOC_ShowWMCursor( _THIS, WMcursor *cursor );

/* Internal helper functions */
static int GetBpp( TDisplayMode displaymode );

/* !! Table for fast conversion from 8 bit to 12/16 bit */
static TUint16 EPOC_HWPalette_256_to_DisplayMode[256];

VideoBootStrap EPOC_bootstrap =
{
	"epoc",
    "EPOC system",
    EPOC_Available,
    EPOC_CreateDevice
};

const TUint32 WindowClientHandle = 9210; //!!

/* Epoc video driver bootstrap functions */

static int EPOC_Available(void)
{
    /* Always available */
    return 1;
}

static void EPOC_DeleteDevice( SDL_VideoDevice *device )
{
	free(device->hidden);
	free(device);
}

static SDL_VideoDevice *EPOC_CreateDevice( int /*devindex*/ )
{
	SDL_VideoDevice *device;

	// allocate device structures, we free those on delete
	device = (SDL_VideoDevice *) malloc(sizeof(SDL_VideoDevice));

	if ( device )
    {
		memset(device, 0, (sizeof *device));

        // allocate the private structure
		device->hidden = (struct SDL_PrivateVideoData *)
				malloc( (sizeof *device->hidden) );
	}

	if ( (device == NULL) || (device->hidden == NULL) )
    {
		SDL_OutOfMemory();

		if ( device )
        {
			free(device);
		}

		return(0);
	}

	memset( device->hidden, 0, (sizeof *device->hidden) );

	/* Set the function pointers */
	device->VideoInit = EPOC_VideoInit;
	device->ListModes = EPOC_ListModes;
	device->SetVideoMode = EPOC_SetVideoMode;
	device->SetColors = EPOC_SetColors;
	device->UpdateRects = NULL;
	device->VideoQuit = EPOC_VideoQuit;
	device->AllocHWSurface = EPOC_AllocHWSurface;
	device->CheckHWBlit = NULL;
	device->FillHWRect = NULL;
	device->SetHWColorKey = NULL;
	device->SetHWAlpha = NULL;
	device->LockHWSurface = EPOC_LockHWSurface;
	device->UnlockHWSurface = EPOC_UnlockHWSurface;
	device->FlipHWSurface = EPOC_FlipHWSurface;
	device->FreeHWSurface = EPOC_FreeHWSurface;
	device->SetIcon = NULL;
	device->SetCaption = NULL;
	device->GetWMInfo = NULL;
	device->FreeWMCursor = EPOC_FreeWMCursor;
	device->CreateWMCursor = EPOC_CreateWMCursor;
	device->ShowWMCursor = EPOC_ShowWMCursor;
	device->WarpWMCursor = NULL;
	device->InitOSKeymap = EPOC_InitOSKeymap;
	device->PumpEvents = EPOC_PumpEvents;
	device->free = EPOC_DeleteDevice;

	return device;
}

int EPOC_VideoInit( _THIS, SDL_PixelFormat *vformat )
{
    // !!TODO:handle leave functions!

	Private->iEikEnv = CEikonEnv::Static();

	Private->iWindowCreator = 
        static_cast<CEBasicAppUi*>( Private->iEikEnv->EikAppUi() )->View();

	/* Initialize all variables that we clean on shutdown */

	for ( int i=0; i < SDL_NUMMODES; ++i )
    {
		Private->SDL_modelist[i] = (SDL_Rect *) malloc( sizeof(SDL_Rect) );

        if (Private->SDL_modelist[i] != NULL)
        {
		    Private->SDL_modelist[i]->x = 0;
            Private->SDL_modelist[i]->y = 0;
            Private->SDL_modelist[i]->w = 0;
            Private->SDL_modelist[i]->h = 0;
        }
        else
        {
            SDL_OutOfMemory();

            for ( int j = 0; j < i; ++j )
            {
                free(Private->SDL_modelist[j]);
            }

            return -1;
        }
    }

	/* Modes sorted largest to smallest !!TODO:sorting order??*/
	Private->SDL_modelist[0]->w = 320; Private->SDL_modelist[0]->h = 200;
    Private->SDL_modelist[1]->w = 320; Private->SDL_modelist[1]->h = 240;
	Private->SDL_modelist[2]->w = 640; Private->SDL_modelist[2]->h = 200;
	Private->SDL_modelist[3]->w = 640; Private->SDL_modelist[3]->h = 400; 
	Private->SDL_modelist[4]->w = 640; Private->SDL_modelist[4]->h = 480;
	Private->SDL_modelist[5] = NULL;

    /* Initialise Epoc frame buffer */

    TDisplayMode displayMode = 
        Private->iEikEnv->ScreenDevice()->DisplayMode();

    #ifndef __WINS__

	TInt value;
	HAL::Get(HALData::EMachineUid,value);
	TInt offset=0;

    // TODO: this whole thing is not very effective because
    // new devices will come out all the time. gotta think
    // of something better to do here.

	switch( value )
	{
	case 0x101fb3dd: // nokia 6600
	case 0x101f466a: // nokia 3650
	case 0x101f4fc3: // nokia 7650
	case 0x101f8c19: // ngage
	case 0x101FB2B1: // ngage qd
	case 0x101f6b26: //A9xx
	case 0x101fd279://P3x
	 offset = 32;
	 break;

	case 0x101f408b://P800
	case 0x101fb2ae://P900
        // no offset for sonyericssons
        offset = 0;
		break;

	default:
        // set the offset to 32 anyway on default
        offset = 32;
		break;
	}

    TScreenInfoV01 screenInfo;
	TPckg<TScreenInfoV01> sInfo(screenInfo);
	UserSvr::ScreenInfo(sInfo);
	
	Private->EPOC_ScreenSize		= screenInfo.iScreenSize; 
	Private->EPOC_DisplayMode		= displayMode;
    Private->EPOC_HasFrameBuffer	= screenInfo.iScreenAddressValid;
	Private->EPOC_FrameBuffer		= Private->EPOC_HasFrameBuffer ? (TUint8*) screenInfo.iScreenAddress+offset : NULL;
	Private->EPOC_BytesPerPixel	    = ((GetBpp(displayMode)-1) / 8) + 1;
	Private->EPOC_BytesPerScanLine	= screenInfo.iScreenSize.iWidth * Private->EPOC_BytesPerPixel;

    /* It seems that in SA1100 machines for 8bpp displays there is a 512 palette table at the 
     * beginning of the frame buffer. E.g. Series 7 and Netbook.
     * In 12 bpp machines the table has 16 entries.
	 */
	if (Private->EPOC_HasFrameBuffer && GetBpp(displayMode) == 8)
		Private->EPOC_FrameBuffer += 512;
	// this is NOT valid for a P800. It has NO palette entry
	//if (Private->EPOC_HasFrameBuffer && GetBpp(displayMode) == 12 || Private->EPOC_HasFrameBuffer && GetBpp(displayMode) == 16)
	//	Private->EPOC_FrameBuffer += 16 * 2;

    #else /* defined __WINS__ */
    
    /* Create bitmap, device and context for screen drawing */
    Private->EPOC_ScreenSize        = Private->iEikEnv->ScreenDevice()->SizeInPixels();

	Private->EPOC_Bitmap = new (ELeave) CWsBitmap(Private->iEikEnv->WsSession());
	Private->EPOC_Bitmap->Create(Private->EPOC_ScreenSize, displayMode);

	Private->EPOC_DisplayMode	    = displayMode;
    Private->EPOC_HasFrameBuffer    = ETrue;
    Private->EPOC_FrameBuffer       = NULL; /* Private->EPOC_Bitmap->DataAddress() can change any time */
	Private->EPOC_BytesPerPixel	    = ((GetBpp(displayMode)-1) / 8) + 1;
	Private->EPOC_BytesPerScanLine  = Private->iEikEnv->ScreenDevice()->SizeInPixels().iWidth * Private->EPOC_BytesPerPixel;

    #endif /* __WINS__ */

    /* The "best" video format should be returned to caller. */

    vformat->BitsPerPixel       = GetBpp(displayMode) ;
    vformat->BytesPerPixel      = Private->EPOC_BytesPerPixel;

    Private->EPOC_ScreenOffset = 0;

    // initialize stuff for mouse emulation
	Private->EPOC_EmulMouseX = 0;
	Private->EPOC_EmulMouseY = 0;
	Private->EPOC_EmulButtonPressed = 0;

    //!! TODO: error handling
	 return(0);
}


SDL_Rect **EPOC_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
{
    if (format->BitsPerPixel == GetBpp(Private->EPOC_DisplayMode)  || format->BitsPerPixel == 8)
        return Private->SDL_modelist;
    return NULL;
}

int EPOC_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors)
{
    for(int i = firstcolor; i < ncolors; i++) {

	    // 4k value: 000rgb
	    TUint16 colornK = 0;
		if(GetBpp(Private->EPOC_DisplayMode)==12)
		{
        colornK |= (colors[i].r & 0x0000f0) << 4; 
	    colornK |= (colors[i].g & 0x0000f0);      
	    colornK |= (colors[i].b & 0x0000f0) >> 4;
		}
		else
		{
			colornK |= (colors[i].r & 0x0000f8) << 8; 
			colornK |= (colors[i].g & 0x0000fc)<<3;      
			colornK |= (colors[i].b & 0x0000f8) >> 3;
		}
        EPOC_HWPalette_256_to_DisplayMode[i] = colornK;
    }
	return(0);
}

void ClearWindowL(_THIS)
{
	TBool lockedHeap=EFalse;
#ifdef __WINS__
	//TBitmapUtil lock(Private->EPOC_Bitmap);	
	//lock.Begin(TPoint(0,0)); // Lock bitmap heap
	if(!gHeapIsLocked)
	{
		Private->EPOC_Bitmap->LockHeap();
		gHeapIsLocked=ETrue;
		lockedHeap=ETrue;
	}
	Private->iWindowCreator->ActGc();
#endif
	
	/* Draw black background */
#ifdef __WINS__
	TUint16* screenBuffer = (TUint16*)Private->EPOC_Bitmap->DataAddress();
#else
	TUint16* screenBuffer = (TUint16*)Private->EPOC_FrameBuffer;
#endif
	for (int y=0; y < Private->EPOC_ScreenSize.iHeight; y++) {
		for (int x=0; x < Private->EPOC_ScreenSize.iWidth; x++) {
			*screenBuffer++ = 0x0000;
		}
	}
	
    /* Tell the system that something has been drawn */
	TRect  rect = TRect(Private->iWindowCreator->Size());
	Private->iWindowCreator->Win().Invalidate(rect);
	
#ifdef __WINS__
	Private->iWindowCreator->Win().BeginRedraw(rect);
	Private->iEikEnv->SystemGc().BitBlt(TPoint(), Private->EPOC_Bitmap);
	Private->iWindowCreator->Win().EndRedraw();
	Private->iWindowCreator->DeGc();
    //lock.End(); // Unlock bitmap heap
	if(lockedHeap)
	{
		Private->EPOC_Bitmap->UnlockHeap();
		gHeapIsLocked=EFalse;
	}
	Private->iEikEnv->WsSession().Flush();
#endif
}

SDL_Surface *EPOC_SetVideoMode(_THIS, SDL_Surface *current,
				int width, int height, int bpp, Uint32 flags)
{
    if (!(((width == 640 && height == 200 ) ||
          (width == 640 && height == 400 ) || 
          (width == 640 && height == 480 ) || 
          (width == 320 && height == 200 )||
		  (width == 320 && height == 240 ) 
		  ) &&(bpp == 8 || bpp == GetBpp(Private->EPOC_DisplayMode) )
		  )){
		SDL_SetError("Requested video mode is not supported");
        return NULL;
    }

    if (current && current->pixels) {
        free(current->pixels);
        current->pixels = NULL;
    }

	if ( ! SDL_ReallocFormat(current, bpp, 0, 0, 0, 0) ) {
		return(NULL);
	}

    /* Set up the new mode framebuffer */
    if (bpp == 8) 
	    current->flags = (SDL_FULLSCREEN|SDL_SWSURFACE|SDL_PREALLOC|SDL_HWPALETTE); 
    else // 12 bpp
	    current->flags = (SDL_FULLSCREEN|SDL_SWSURFACE|SDL_PREALLOC); 
	current->w = width;
	current->h = height;
    int numBytesPerPixel = ((bpp-1)>>3) + 1;   
	current->pitch = numBytesPerPixel * width; // Number of bytes in scanline 

	current->pixels = malloc(width * height * numBytesPerPixel);
	if(current->pixels == NULL)
	{
		// free current?
		return(NULL);
	}
	memset(current->pixels, 0, width * height * numBytesPerPixel);

    Private->EPOC_RotateMode = EPOC_ROTATE_CLOCKWISE;

	int add = 1;
	int modulo = 1;
	
	TSize scaledSize = TSize(width, height);
	
	TSize dstSize;
	
	// Fix rotation settings to new "virtual" screen dimensions

	if(Private->EPOC_RotateMode != EPOC_ROTATE_NONE)
	{
		dstSize = TSize(Private->EPOC_ScreenSize.iHeight, Private->EPOC_ScreenSize.iWidth);
	}
	else
	{
		dstSize = TSize(Private->EPOC_ScreenSize.iWidth, Private->EPOC_ScreenSize.iHeight);
	}
	
	if (dstSize.iWidth < width || dstSize.iHeight < height)
	{
		if (height * dstSize.iWidth < dstSize.iHeight * width)
		{
			// scale is determined by horizontal scale
			add = dstSize.iWidth;
			modulo = width;
			scaledSize.iWidth = dstSize.iWidth;
			scaledSize.iHeight =
				(height * dstSize.iWidth + width - 1) / width;
		}
		else
		{
			// scale is determined by vertical scale
			add = dstSize.iHeight;
			modulo = height;
			scaledSize.iWidth =
				(width * dstSize.iHeight + height - 1) / height;
			scaledSize.iHeight = dstSize.iHeight;
		}
	}
	
	Private->EPOC_StartPos = TPoint((dstSize.iWidth - scaledSize.iWidth) / 2,
		(dstSize.iHeight - scaledSize.iHeight) / 2);
	
	Private->EPOC_Add = add;
	Private->EPOC_Modulo = modulo;

	/* Set the blit function */
	_this->UpdateRects = EPOC_DirectUpdate;

	ClearWindowL(_this);
    /* Centralize game window on device screen  */
    Private->EPOC_ScreenOffset = 0;

	return(current);
}

void RedrawWindowL(_THIS)
{
    SDL_Rect fullScreen;
    fullScreen.x = 0;
    fullScreen.y = 0;
    fullScreen.w = _this->screen->w;
    fullScreen.h = _this->screen->h;
	TBool lockedHeap=EFalse;
#ifdef __WINS__
	    //TBitmapUtil lock(Private->EPOC_Bitmap);	
        //lock.Begin(TPoint(0,0)); // Lock bitmap heap
		if(!gHeapIsLocked)
		{
			Private->EPOC_Bitmap->LockHeap();
			gHeapIsLocked=ETrue;
			lockedHeap=ETrue;
		}
	    Private->iWindowCreator->ActGc();
#endif

    if (fullScreen.w < Private->EPOC_ScreenSize.iWidth
        && fullScreen.w < Private->EPOC_ScreenSize.iWidth) {
        /* Draw blue stripes background */
#ifdef __WINS__
        TUint16* screenBuffer = (TUint16*)Private->EPOC_Bitmap->DataAddress();
#else
        TUint16* screenBuffer = (TUint16*)Private->EPOC_FrameBuffer;
#endif
        for (int y=0; y < Private->EPOC_ScreenSize.iHeight; y++) {
            for (int x=0; x < Private->EPOC_ScreenSize.iWidth; x++) {
                TUint16 color = ((x+y)>>1) & 0xf; /* Draw pattern */
                *screenBuffer++ = color;
            }
        }
    }

    /* Tell the system that something has been drawn */
	TRect  rect = TRect(Private->iWindowCreator->Size());
  	Private->iWindowCreator->Win().Invalidate(rect);

#ifdef __WINS__
	Private->iWindowCreator->Win().BeginRedraw(rect);
	Private->iEikEnv->SystemGc().BitBlt(TPoint(), Private->EPOC_Bitmap);
	Private->iWindowCreator->Win().EndRedraw();
	Private->iWindowCreator->DeGc();
    //lock.End(); // Unlock bitmap heap
	if(lockedHeap)
		{
			Private->EPOC_Bitmap->UnlockHeap();
			gHeapIsLocked=EFalse;
		}
	Private->iEikEnv->WsSession().Flush();
#endif

    /* Draw current buffer */
    EPOC_DirectUpdate(_this, 1, &fullScreen);
}


/* We don't actually allow hardware surfaces other than the main one */
static int EPOC_AllocHWSurface(_THIS, SDL_Surface *surface)
{
	return(-1);
}
static void EPOC_FreeHWSurface(_THIS, SDL_Surface *surface)
{
	return;
}

static int EPOC_LockHWSurface(_THIS, SDL_Surface *surface)
{
	return(0);
}
static void EPOC_UnlockHWSurface(_THIS, SDL_Surface *surface)
{
	return;
}

static int EPOC_FlipHWSurface(_THIS, SDL_Surface *surface)
{
	return(0);
}

#define COMPUTE_RGB_AVERAGE(a, b, mask) ((((a)&(mask))>>1)+(((b)&(mask))>>1)+((a)&(~(mask))))

void
DrawScaled16BitMameLineTo12or16Bpp(
  TUint16 *src_start,
  int add,
  int modulo,
  TUint16 *dst,
  int visibleColumns,
  int lenToNextPixel,
  TInt mask)
{

  TInt accumulator = 0;

  TUint16 *src = src_start;

  TInt pixel = 0;
  for (TInt w = visibleColumns - 1; w >= 0; w--)
  {
    TUint srcPixel = *src;//pal[*src];
    src++;
    if (accumulator >= add)
    {
      srcPixel = COMPUTE_RGB_AVERAGE(pixel, srcPixel, mask);
    }
    pixel = srcPixel;
    
    accumulator += add;
    if (accumulator >= modulo || w == 0)
    {
      *dst = pixel;
//      dst++;
	  dst += lenToNextPixel;
      accumulator -= modulo;
    }
  }
}

void
DrawScaled16BitMameLineTo12or16BppWithMerge(
  TUint16 *src_start,
  int add,
  int modulo,
  TUint16 *dst,
  int visibleColumns,
  int lenToNextPixel,
  TInt mask)
{

  TInt accumulator = 0;

  TUint16 *src = src_start;

  TInt pixel = 0;
  for (TInt w = visibleColumns - 1; w >= 0; w--)
  {
    TUint srcPixel = *src;//pal[*src];
    src++;
    if (accumulator >= add)
    {
      srcPixel = COMPUTE_RGB_AVERAGE(pixel, srcPixel, mask);
    }
    pixel = srcPixel;
    
    accumulator += add;
    if (accumulator >= modulo || w == 0)
    {
      pixel = COMPUTE_RGB_AVERAGE(*dst, pixel, mask); // merge
      *dst = pixel;
	  dst += lenToNextPixel;
      accumulator -= modulo;
    }
  }
}

static void EPOC_DirectUpdate(_THIS, int numrects, SDL_Rect *rects)
{
	TInt mask = 0; // Clears lsb for each color component
	switch (GetBpp(Private->EPOC_DisplayMode))
	{
    case 12:
		mask = 0x0eee; // 0000 bbbb gggg rrrr : 0000 1110 1110 1110
		break;
    case 16:
		mask = 0xf7de; // bbbb bggg gggr rrrr : 1111 0111 1101 1110
		break;
    default:
		ASSERT(EFalse);
		break;
	}
	
	TUint16 *src = (TUint16 *)_this->screen->pixels;
	int srcAddPerRow = _this->screen->w;
	
	int dstAddPerPixel;
	int dstAddPerRow;
	
	TInt srcLines = _this->screen->h;
	TInt srcColumns = _this->screen->w;
	TInt accumulator = 0;
	TBool merge = EFalse;
	TInt add = Private->EPOC_Add;
	TInt modulo = Private->EPOC_Modulo;
	TBool lockedHeap = EFalse;
	
#ifdef __WINS__
	if(!gHeapIsLocked)
	{
		Private->EPOC_Bitmap->LockHeap();
		gHeapIsLocked=ETrue;
		lockedHeap=ETrue;
	}
	Private->iWindowCreator->ActGc();
	TUint16* dst = (TUint16*)Private->EPOC_Bitmap->DataAddress();
#else
	TUint16* dst = (TUint16*)Private->EPOC_FrameBuffer;
#endif
	
	if (Private->EPOC_RotateMode == EPOC_ROTATE_CLOCKWISE)
	{
		dstAddPerRow = -1;
		dstAddPerPixel = Private->EPOC_ScreenSize.iWidth;
		// determines where our 0,0 point is (e.g 176,0)
		dst += (Private->EPOC_ScreenSize.iWidth - 1);
	}
    else if(Private->EPOC_RotateMode == EPOC_ROTATE_COUNTERCLOCKWISE)
    {
        dstAddPerPixel = -Private->EPOC_ScreenSize.iWidth;
        dstAddPerRow = 1;
        // determines where our 0,0 point is (e.g 0,208)
        dst += (Private->EPOC_ScreenSize.iHeight - 1) * Private->EPOC_ScreenSize.iWidth;
    }
	else
	{
		dstAddPerRow = Private->EPOC_ScreenSize.iWidth;
		dstAddPerPixel = 1;
        // if there's no rotation, there's no speical 0,0 point
	}

    // now add the offset of the starting position
    dst += (Private->EPOC_StartPos.iY * dstAddPerRow) // Y
		+ (Private->EPOC_StartPos.iX * dstAddPerPixel); // X
	
	for (TInt h = 0; h < srcLines; h++)
	{
		if (merge)
		{
			DrawScaled16BitMameLineTo12or16BppWithMerge(src,
				add,
				modulo,
				dst,
				srcColumns,
				dstAddPerPixel,
				mask);
		}
		else
		{
			DrawScaled16BitMameLineTo12or16Bpp(src,
				add,
				modulo,
				dst,
				srcColumns,
				dstAddPerPixel,
				mask);
		}
		merge = ETrue;
		accumulator += add;
		if (accumulator >= modulo || h == srcLines - 1)
		{
			dst += dstAddPerRow;
			merge = EFalse;
			accumulator -= modulo;
		}
		src += srcAddPerRow;
	}
	
#ifdef __WINS__
	TRect  myrect = TRect(Private->iWindowCreator->Size());
	Private->iWindowCreator->Win().Invalidate(myrect);
	Private->iWindowCreator->Win().BeginRedraw(myrect);
	Private->iEikEnv->SystemGc().BitBlt(TPoint(), Private->EPOC_Bitmap);
	Private->iWindowCreator->Win().EndRedraw();
	Private->iWindowCreator->DeGc();
	//    lock.End(); // Unlock bitmap heap
	if(lockedHeap)
	{
		Private->EPOC_Bitmap->UnlockHeap();
		gHeapIsLocked=EFalse;
	}
	Private->iEikEnv->WsSession().Flush();
#endif
	
#ifdef SYMBIAN_SERIES_60
	TRawEvent redrawEvent;
	redrawEvent.Set(TRawEvent::ERedraw);
	UserSvr::AddEvent(redrawEvent);
#endif
	
}

/* Note:  If we are terminated, this could be called in the middle of
   another SDL video routine -- notably UpdateRects.
*/
void EPOC_VideoQuit(_THIS)
{
	int i;

	/* Free video mode lists */
	for ( i=0; i<SDL_NUMMODES; ++i ) {
		if ( Private->SDL_modelist[i] != NULL ) {
			free(Private->SDL_modelist[i]);
			Private->SDL_modelist[i] = NULL;
		}
	}
	
    if ( _this->screen && (_this->screen->flags & SDL_HWSURFACE) ) {
		/* Direct screen access, no memory buffer */
		_this->screen->pixels = NULL;
	}

    if (_this->screen && _this->screen->pixels) {
        free(_this->screen->pixels);
        _this->screen->pixels = NULL;
    }

    /* Free Epoc resources */

    /* Disable events for me */

    #ifdef __WINS__
	delete Private->EPOC_Bitmap;
	Private->EPOC_Bitmap = NULL;
    #endif
}

int GetBpp( TDisplayMode displaymode )
{
    TInt numColors = TDisplayModeUtils::NumDisplayModeColors( displaymode );
    TInt bitsPerPixel = 1;

    for ( TInt32 i = 2; i < numColors; i <<= 1, bitsPerPixel++ );

    return bitsPerPixel;    
}

WMcursor *EPOC_CreateWMCursor(_THIS, Uint8 *data, Uint8 *mask, int w, int h, int hot_x, int hot_y)
{
	return (WMcursor *) 9210; // it's ok to return something unuseful but true
}

void EPOC_FreeWMCursor(_THIS, WMcursor *cursor)
{
    /* Disable virtual cursor */
    HAL::Set(HAL::EMouseState, HAL::EMouseState_Invisible);
    Private->iEikEnv->WsSession().SetPointerCursorMode(EPointerCursorNone);
}

int EPOC_ShowWMCursor(_THIS, WMcursor *cursor)
{

    if (cursor ==  (WMcursor *)9210) {
        /* Enable virtual cursor */
	    HAL::Set(HAL::EMouseState, HAL::EMouseState_Visible);
	    Private->iEikEnv->WsSession().SetPointerCursorMode(EPointerCursorNormal);
    }
    else {
        /* Disable virtual cursor */
        HAL::Set(HAL::EMouseState, HAL::EMouseState_Invisible);
       Private->iEikEnv->WsSession().SetPointerCursorMode(EPointerCursorNone);
    }

	return(1);
}

}; // extern "C"

